import os
from . import base


# Yes, all of these overlayed methods have to exist for IDE features to work,
# one does not simply type hint an inherited variable


class Toggle(base.Setting):
	"""A boolean True/False"""
	def __init__(self, name, default: bool, section=base.default_section):
		super().__init__(bool, name, default, section)


	def get(self) -> bool:
		return super().get()

	def set(self, new_value: bool):
		super().set(new_value)


	@property
	def value(self) -> bool:
		return super().value

	@value.setter
	def value(self, new_value: bool):
		super().value = new_value


	def toggle(self):
		self.set(not self.get())

	def __bool__(self):
		return self.value





class Choice(base.Setting):
	"""Choose an option (str) from a list"""
	def __init__(self, name, options: list[str], default: str, section=base.default_section):
		self.options = options
		super().__init__(str, name, default, section)

	def validate(self, value):
		return super().validate(value) and (value in self.options)

	def get(self) -> str:
		return super().get()

	def set(self, new_value: str):
		super().set(new_value)


	@property
	def value(self) -> str:
		return super().value

	@value.setter
	def value(self, new_value: str):
		super().value = new_value





class Multichoice(base.Setting):
	"""Choose multiple options (str) from a list"""
	def __init__(self, name, options: list[str], default_choices: list[str], section=base.default_section):
		self.options = options
		super().__init__(list, name, default_choices, section)


	def validate(self, value):
		conditions = [
			super().validate(value),
			all((item in self.options) for item in value),
			all((type(item) == str) for item in value)
		]
		return all(*conditions)


	def get(self) -> list[str]:
		return super().get()

	def set(self, new_value: list[str]):


		for value in new_value:
			if value not in self.options:
				raise ValueError(f"New value {new_value} not in options ({self.options})")

		super().set(new_value)


	@property
	def value(self) -> list[str]:
		return super().value

	@value.setter
	def value(self, new_value: list[str]):
		super().value = new_value


	def append(self, item: str):
		value = self.get()
		value.append(item)
		self.set(value)

	def remove(self, item: str):
		value = self.get()
		value.remove(item)
		self.set(value)





class Text(base.Setting):
	"""Just normal text"""
	def __init__(self, name, default: str, section=base.default_section):
		super().__init__(str, name, default, section)

	def get(self) -> str:
		return super().get()

	def set(self, new_value: str):
		super().set(new_value)

	@property
	def value(self) -> str:
		return super().value

	@value.setter
	def value(self, new_value: str):
		super().value = new_value



class Path(base.Setting):
	"""A file path. Automatically converted between Windows and Unix paths."""
	def __init__(self, name, default: str, has_to_exist: bool = False, section=base.default_section):
		self.has_to_exist = has_to_exist
		super().__init__(str, name, default, section)


	def validate(self, value):
		valid = super().validate(value)
		if self.has_to_exist:
			return valid and os.path.exists(value)
		else:
			return valid


	def get(self) -> str:
		return super().get()

	def set(self, new_value: str):
		if os.name == 'nt':
			new_value = new_value.replace('/', '\\')
		else:
			new_value = new_value.replace('\\', '/')
		super().set(new_value)

	@property
	def value(self) -> str:
		return super().value

	@value.setter
	def value(self, new_value: str):
		super().value = new_value




class Number(base.Setting):
	"""A number (int) that can be incremented and decremented"""
	def __init__(self, name, default: int, lower_limit: int = 0, upper_limit: int = 100,
	             section=base.default_section):
		self.lower_limit = lower_limit
		self.upper_limit = upper_limit
		super().__init__(int, name, default, section)


	def validate(self, value):
		if type(value) == int:  # The base method does check for this, but the size check could crash
			return super().validate(value) and (self.lower_limit <= value <= self.upper_limit)
		else:
			return False


	def get(self) -> int:
		return super().get()

	def set(self, new_value: int):
		super().set(new_value)

	def increment(self):
		self.set(self.get() + 1)

	def decrement(self):
		self.set(self.get() - 1)



import appdirs, os


_dir: str | None = None


def setup(app_name, custom_dir: str = None):
	"""Creates the config dir with your app name. Required to use the package."""
	if isinstance(custom_dir, str):
		config_dir = custom_dir
	else:
		# noinspection PyTypeChecker
		config_dir = appdirs.user_config_dir(app_name, False)
	os.makedirs(config_dir, exist_ok=True)
	global _dir
	_dir = config_dir


def get():
	"""Return where the config values are being stored"""
	if isinstance(_dir, str):
		return _dir
	else:
		raise NotSetupError


class NotSetupError(BaseException):
	"""Called when no app name has been set up"""
	def __init__(self):
		super().__init__("You haven't set your app name (required for deciding the config file directory). "
						 "Call setup() with the name.")



"""
Simple-to-use object-oriented config library.

The values have automatic validation and get saved to a file that's
seamlessly written and read in the background, so you don't have to worry about any of it.
"""

from . import base, dir, parser, types

from .dir import setup
from .base import Section, all_sections
from .types import *




import os

from . import dir, parser




class Section:
	"""A group of settings that is also represented by one file.
	Usage of these is optional, settings defined without sections will use the default one."""
	def __init__(self, name: str):
		self.name = name
		self.settings = []
		self.file = parser.ConfigFile(self.get_path, title=self.name)
		all_sections.append(self)

	def get_path(self):
		return os.path.join(dir.get(), self.name + '.cfg')



all_sections: list[Section] = []
default_section = Section('Settings')




class Setting:
	"""Base setting class"""
	def __init__(self, datatype, name: str, default, section: Section = default_section):
		self.datatype = datatype
		self.name = name
		self.default = default
		self.section = section

		if self not in self.section.settings:
			self.section.settings.append(self)

		# Make sure that option exists in fle
		if self.name not in self.section.file.keys():
			self.set(self.default)


	def validate(self, value) -> bool:
		return type(value) == self.datatype
		# Expanded within each setting type



	def get(self):
		"""Return stored value, or the default if invalid or missing"""
		value = self.section.file.get(self.name, self.datatype)
		if self.validate(value):
			return value
		else:
			return self.default

	def set(self, new_value):
		"""Validate and set a new value. Invalid values will raise a ValueError."""
		valid = self.validate(new_value)
		if not valid:
			raise ValueError(f"New setting value {new_value} is invalid")
		self.section.file.set(self.name, new_value)


	@property
	def value(self):
		return self.get()


	@value.setter
	def value(self, new_value):
		self.set(new_value)


	def __repr__(self):
		return f"{self.name}: {self.value}"

	def __iter__(self):
		return self.value.__iter__()


	def __eq__(self, other):
		"""Supports comparing with actual types (and other settings)"""
		if isinstance(other, Setting):
			return (other.name == self.name) and (other.section == self.section)
		elif isinstance(other, self.datatype):
			return self.value == other




import configparser
import os.path
import typing


class ConfigFile:
	"""Controller/parser for a config file.
	Can technically be used by itself without the rest of the library, if you want dict-style settings instead."""
	def __init__(self, path: str | typing.Callable[[], str], title="Settings"):
		"""
		Controller/parser for a config file.
		Can technically be used by itself without the rest of the library, if you want dict-style settings instead.

		:param path: Either a path string or a function that returns a path string. File doesn't have to exist in advance.
		:param title: Header to write to the top of the file
		"""
		super().__init__()
		self._path = path
		self.title = title

		# Initialize lower-level config parser
		self._parser = configparser.ConfigParser()
		self._parser.optionxform = str  # For preserve case

		# Make sure that section exists
		if self.title not in self._parser.sections():
			self._parser.add_section(self.title)

	@property
	def path(self):
		"""Get the path of the file."""
		if callable(self._path):  # For the default section, so it can be defined without knowing the dir yet
			return str(self._path())
		else:
			return str(self._path)


	def reload(self):
		"""Load values from file. If file is invalid it will be ignored and overwritten."""
		try:
			self._parser.read(self.path)
		except:
			self._write()

	def _write(self):
		"""Dump all current values to the file."""
		os.makedirs(os.path.dirname(self.path), exist_ok=True)
		with open(self.path, 'w') as f:
			self._parser.write(f)



	def get(self, option: str, type: typing.Type[str | int | float | bool | list]):
		"""Reload and get value."""
		self.reload()
		special_functions = {
			str: self._parser.get,
			int: self._parser.getint,
			float: self._parser.getfloat,
			bool: self._parser.getboolean,
			list: lambda *args: self._parser.get(*args).strip("['").strip("']").split("', '")
		}
		function = special_functions[type]
		return function(self.title, option)


	def set(self, option: str, value):
		"""Set a value and write to the file."""
		self._parser.set(self.title, option, str(value))
		self._write()


	def remove(self, option: str):
		"""Remove an option (and update the file)."""
		self._parser.remove_option(self.title, option)
		self._write()


	def keys(self):
		"""Return list of all defined options."""
		self.reload()
		return self._parser.options(self.title)
